////////////////////////////////////////////////////////////////////////
// DesktopFolders: a FolderView for your background desktop
//
// Some ideas taken from FolderView screenlet, by Helder Fraga (Whise) and Julien Lavergne
//
// Requirements: gtk+-3, libgee (run) and vala (build)
//
// To compile:
//    $ valac --pkg gtk+-3.0 --pkg cairo --pkg gee-1.0  -X -O3  DesktopFolders.vala
//
// Copyright (C) 2011  Iñigo Serna <inigoserna@gmail.com>
// Released under GPL v3+
//
// Time-stamp: <Sun Oct  9 15:52:16 2011>
//////////////////////////////////////////////////////////////////////


using Gdk;
using Gtk;
using Cairo;
using Pango;

using Gee;


///////////////////////////////////////////////////////////////////////////
///// Constants
const string[] PROGRAM_AUTHORS = { "Iñigo Serna <inigoserna@gmail.com>", null };
const string   PROGRAM_NAME    = "Desktop Folders";
const string   PROGRAM_VERSION = "1.0";
const string   PROGRAM_WEBSITE = "http://inigo.katxi.org/devel/DesktopFolders";

const string CONFIG_FILE = ".config/DesktopFolders/config.ini";
const uint   CONFIG_SAVING_TIMEOUT = 3600; // 1 hour

const int    DEFAULT_WIN_X0      = 100;
const int    DEFAULT_WIN_Y0      = 100;
const int    DEFAULT_WIN_WIDTH   = 550;
const int    DEFAULT_WIN_HEIGHT  = 450;
const int    DEFAULT_NCOLS       = 5;
const string DEFAULT_HEADER_FONT = "Fertigo Pro 15";
const int    DEFAULT_ICON_SIZE   = 60;
const string DEFAULT_LABEL_FONT  = "Segoe UI 11";

const string[] LIST_ICON_SIZE_STR  = { "Tiny", "Small", "Medium", "Big", "Large", "Huge" };
const int[]    LIST_ICON_SIZE      = { 36, 48, 60, 72, 96, 128 };
const int[]    LIST_ITEM_WIDTH     = { 68, 80, 92, 100, 118, 148 };
const int BORDER          = 5;
const int SEP_PADDING     = 6;
const int SEP_SIZE        = 1;
const int ITEM_PADDING    = 7;
const int ITEM_SPACING    = 2;

///// Global variables
ArrayList<DesktopFolder> dfs;
ThumbFactory thumb_factory;


///////////////////////////////////////////////////////////////////////////
///// DFItem
public class DFItem: GLib.Object {
    public string     parent_uri;
    public FileInfo   fileinfo;
    public string     name;
    public bool       isdir;
    public bool       islink;
    public string     type;
    public int64      size;
    public DFItemUI   ui;

    public DFItem(FileInfo fi, File parent) throws GLib.Error {
        this.parent_uri = parent.get_uri();
        this.fileinfo = fi;
        this.name = fi.get_name();
        this.isdir = fi.get_file_type() == FileType.DIRECTORY;
        this.islink = fi.get_is_symlink();
        this.ui = new DFItemUI(this);
    }
}


public class DFItemUI {
    private DFItem df;
    public int x0;
    public int y0;
    public int x1;
    public int y1;
    public int r;
    public int c;
    public Gdk.Pixbuf thumb;
    public int        thumb_x;
    public int        thumb_y;
    public string     label;
    public int        label_x;
    public int        label_y;

    public DFItemUI(DFItem df) {
        this.df = df;
    }

    public void fill_geometry(int x0, int y0, int x1, int y1, int r, int c, DFConfig cfg) {
        this.x0 = x0;
        this.y0 = y0;
        this.x1 = x1;
        this.y1 = y1;
        this.r = r;
        this.c = c;
        this.thumb = thumb_factory.get_thumb(this.df, cfg.ui_icon_size);
        if (this.thumb != null) {
            this.thumb_x = (x0) + (cfg.ui_item_width-thumb.width)/2;
            this.thumb_y = (y0+ITEM_PADDING) + (cfg.ui_icon_size-thumb.height)/2;
        } else {
            this.thumb_x = 0;
            this.thumb_y = 0;
        }
        this.label = null;		// to be filled later
        this.label_x = -1;		// to be filled later
        this.label_y = this.y1 - ITEM_PADDING; // = y0+ITEM_PADDING+ITEM_ICONSIZE+2+ITEM_FONTSIZE;
    }

    public void fill_extra(string label, int label_x) {
        this.label = label;
        this.label_x = label_x;
    }

    public bool is_inside(int x, int y) {
        return (x>=x0 && x<=x1 && y>=y0 && y<= y1);
    }
}


public ArrayList<DFItem> get_items(File uri, Filter filter) throws GLib.Error {
    var items = new ArrayList<DFItem>();
    var enumerator = uri.enumerate_children("*", FileQueryInfoFlags.NONE);
    FileInfo fi;
    while ((fi = enumerator.next_file()) != null) {
        if (filter.is_ok(fi))
            items.add(new DFItem(fi, uri));
    }
    items.sort((e1, e2) => {
            var a = (DFItem) e1;
            var b = (DFItem) e2;
            if (a.isdir && !b.isdir)
                return -1;
            if (!a.isdir && b.isdir)
                return 1;
            return GLib.strcmp(a.name.down(), b.name.down());
        });
    return items;
}


string format_size(int64 size, string sep=" ") {
    if (size >= 1000000000)
        return "%.2f%sGB".printf(size/(1024*1024*1024.0), sep);
    else if (size >= 1000000)
        return "%.2f%sMB".printf(size/(1024*1024.0), sep);
    else if (size >= 1000)
       	return "%.2f%sKB".printf(size/1024.0, sep);
    else
        return "%lld%sB".printf(size, sep);
}


///////////////////////////////////////////////////////////////////////////
///// ThumbFactory
public class ThumbFactory : GLib.Object {
    Gtk.IconTheme icontheme;

    public ThumbFactory() {
        this.icontheme = new Gtk.IconTheme().get_default();
        this.icontheme.changed.connect(() => {
                foreach (DesktopFolder df in dfs) {
                    df.update_win_geometry();
                }
            });
    }

    private string? _get_name(string[] iconnames, int iconsize) {
        var iconinfo = icontheme.choose_icon(iconnames, iconsize, Gtk.IconLookupFlags.USE_BUILTIN);
        if (iconinfo != null) {
            return iconinfo.get_filename();
        } else {
            foreach (string icname in iconnames) {
                if (icontheme.has_icon(icname))
                    return icname;
            }
        }
        return null;
    }

    private Gdk.Pixbuf? _get_thumb(DFItem it, int iconsize) throws GLib.Error {
        var name = it.fileinfo.get_attribute_as_string(GLib.FILE_ATTRIBUTE_THUMBNAIL_PATH);
        if (name==null || name=="") {
            var gicon = (ThemedIcon) it.fileinfo.get_icon();
            name = this._get_name(gicon.get_names(), iconsize);
            if (name == null)
                name = this._get_name({"exec"}, iconsize);
        }
        if (name == null)
            return null;
        return new Gdk.Pixbuf.from_file_at_size(name, iconsize, iconsize);
    }

    public Gdk.Pixbuf? get_thumb(DFItem it, int iconsize=DEFAULT_ICON_SIZE) {
        Gdk.Pixbuf thumb;
        try {
            thumb = _get_thumb(it, iconsize);
        } catch (GLib.Error e) {
            thumb = null;
        }
        return thumb;
    }
}


///////////////////////////////////////////////////////////////////////////
///// Filter
public class Filter {
    public bool include_hidden;
    public bool include_backups;
    public bool include_folders;
    public bool include_files;
    public string[] showonly_list;
    public string[] exclude_list;

    public Filter() {
        this.include_hidden = false;
        this.include_backups = false;
        this.include_folders = true;
        this.include_files = true;
        this.showonly_list = {};
        this.exclude_list = {};
    }

    public bool is_ok(FileInfo fi) {
        if (!this.include_hidden && fi.get_is_hidden())
            return false;
        if (!this.include_backups && fi.get_is_backup())
            return false;
        if (!this.include_folders && (fi.get_file_type()==FileType.DIRECTORY))
            return false;
        if (!this.include_files && (fi.get_file_type()!=FileType.DIRECTORY))
            return false;
        var name = fi.get_name();
        foreach (string pat in this.exclude_list) {
            if (pat in name) {
                return false;
            }
        }
        if (this.showonly_list.length > 0) {
            foreach (string pat in this.showonly_list) {
                if (pat in name)
                    return true;
            }
            return false;
        }
        return true;
    }
}


///////////////////////////////////////////////////////////////////////////
///// DFConfig
public class DFConfig : GLib.Object {
    // basic values
    public string uri;
    public Filter filter;
    public int x0;
    public int y0;
    public int w;
    public int h;
    public int ncols;
    public bool   ui_header_fullpath;
    public string ui_header_font;
    public int    ui_icon_size;
    public string ui_label_font;
    // these values are calculated
    public int    nrows;
    public int    ui_item_width;
    public int    ui_item_height;
    public string ui_header_font_name;
    public bool   ui_header_font_italic;
    public bool   ui_header_font_bold;
    public int    ui_header_font_size;
    public string ui_label_font_name;
    public bool   ui_label_font_italic;
    public bool   ui_label_font_bold;
    public int    ui_label_font_size;

    public DFConfig() {
        this.uri = "file://" + Environment.get_home_dir();
        this.x0 = DEFAULT_WIN_X0;
        this.y0 = DEFAULT_WIN_Y0;
        this.w = DEFAULT_WIN_WIDTH;
        this.h = DEFAULT_WIN_HEIGHT;
        this.ncols = DEFAULT_NCOLS;
        this.filter = new Filter();
        this.ui_header_fullpath = true;
        this.ui_header_font = DEFAULT_HEADER_FONT;
        this.ui_icon_size = DEFAULT_ICON_SIZE;
        this.ui_label_font = DEFAULT_LABEL_FONT;
        this.nrows = -1;
        this.update_geo();
    }

    public void update_geo() {
        this._font2attributes(this.ui_header_font, out this.ui_header_font_name,
                              out this.ui_header_font_italic, out this.ui_header_font_bold,
                              out this.ui_header_font_size);
        this._font2attributes(this.ui_label_font, out this.ui_label_font_name,
                              out this.ui_label_font_italic, out this.ui_label_font_bold,
                              out this.ui_label_font_size);
        this.ui_item_width = this._get_item_width();
        this.ui_item_height = 2*ITEM_PADDING + this.ui_icon_size + this.ui_label_font_size + 3;
    }

    private int _get_item_width() {
        var idx = -1;
        for (int i=0; i<LIST_ICON_SIZE.length; i++) {
            if (this.ui_icon_size == LIST_ICON_SIZE[i]) {
                idx = i;
                break;
            }
        }
        return (idx>=0 && idx<LIST_ICON_SIZE.length) ? LIST_ITEM_WIDTH[idx] : DEFAULT_ICON_SIZE;
    }

    private void _font2attributes(string font, out string fontname, out bool is_italic, out bool is_bold, out int fontsize) {
        is_italic = "Italic" in font;
        is_bold = "Bold" in font;
        int idx = font.last_index_of(" ");
        fontsize = int.parse(font[idx:font.length].strip());
        if (fontsize == 0)
            fontsize = 12;
        fontname = font[0:idx].replace("Italic", "").replace("SemiBold", "").replace("Bold", "").strip();
// print("Font: %s -> %s %s %s %d\n", font, fontname, is_italic ? "I" : "i", is_bold ? "B" : "b", fontsize);
    }

    public string to_string(string name) {
        var ui_is_header_fullpath = this.bool2str(this.ui_header_fullpath);
        var f_hidden = this.bool2str(this.filter.include_hidden);
        var f_backups = this.bool2str(this.filter.include_backups);
        var f_folders = this.bool2str(this.filter.include_folders);
        var f_files = this.bool2str(this.filter.include_files);
        var f_showonly = string.joinv(",", this.filter.showonly_list);
        var f_exclude = string.joinv(",", this.filter.exclude_list);
        return @"[$name]\nURI=$uri\nX=$x0\nY=$y0\nWIDTH=$w\nHEIGHT=$h\nNUMBER_COLS=$ncols\nUI_HEADER_FULLPATH=$ui_is_header_fullpath\nUI_HEADER_FONT=$ui_header_font\nUI_ICON_SIZE=$ui_icon_size\nUI_LABEL_FONT=$ui_label_font\nFILTER_INCLUDE_HIDDEN=$f_hidden\nFILTER_INCLUDE_BACKUPS=$f_backups\nFILTER_INCLUDE_FOLDERS=$f_folders\nFILTER_INCLUDE_FILES=$f_files\nFILTER_SHOWONLY_LIST=$f_showonly\nFILTER_EXCLUDE_LIST=$f_exclude\n\n";
    }

    private string bool2str(bool flag) {
        return flag ? "true" : "false";
    }
}


public GLib.List<DFConfig>? config_load() {
    var cfg_path = GLib.Path.build_filename(Environment.get_home_dir(), CONFIG_FILE);
    if (!FileUtils.test(cfg_path, FileTest.IS_REGULAR))
		return null;
	var cfgs = new GLib.List<DFConfig>();
	try {
		var keyfile = new GLib.KeyFile();
		keyfile.load_from_file(cfg_path, KeyFileFlags.NONE);
		foreach (string g in keyfile.get_groups()) {
			if (!("FolderDesktop #" in g))
				continue;
			var cfg = new DFConfig();
			try {
				foreach (string k in keyfile.get_keys(g)) {
					switch (k) {
						case "URI":
							cfg.uri = keyfile.get_string(g, k);
							break;
						case "X":
							cfg.x0 = keyfile.get_integer(g, k);
							break;
						case "Y":
							cfg.y0 = keyfile.get_integer(g, k);
							break;
						case "WIDTH":
							cfg.w = keyfile.get_integer(g, k);
							break;
						case "HEIGHT":
							cfg.h= keyfile.get_integer(g, k);
							break;
						case "NUMBER_COLS":
							cfg.ncols= keyfile.get_integer(g, k);
							break;
						case "UI_HEADER_FULLPATH":
							cfg.ui_header_fullpath= keyfile.get_boolean(g, k);
							break;
						case "UI_HEADER_FONT":
							cfg.ui_header_font= keyfile.get_string(g, k);
							break;
						case "UI_ICON_SIZE":
							cfg.ui_icon_size= keyfile.get_integer(g, k);
							break;
						case "UI_LABEL_FONT":
							cfg.ui_label_font= keyfile.get_string(g, k);
							break;
						case "FILTER_INCLUDE_HIDDEN":
							cfg.filter.include_hidden = keyfile.get_boolean(g, k);
							break;
						case "FILTER_INCLUDE_BACKUPS":
							cfg.filter.include_backups = keyfile.get_boolean(g, k);
							break;
						case "FILTER_INCLUDE_FOLDERS":
							cfg.filter.include_folders = keyfile.get_boolean(g, k);
							break;
						case "FILTER_INCLUDE_FILES":
							cfg.filter.include_files = keyfile.get_boolean(g, k);
							break;
						case "FILTER_SHOWONLY_LIST":
							string[] list = {};
							foreach (var s in keyfile.get_string(g, k).split(",")) {
								list += s;
							}
							cfg.filter.showonly_list = list;
							break;
						case "FILTER_EXCLUDE_LIST":
							string[] list = {};
							foreach (var s in keyfile.get_string(g, k).split(",")) {
								list += s;
							}
							cfg.filter.exclude_list = list;
							break;
						default:
							GLib.stderr.printf("Invalid configuration option: %s in '%s'\n", k, g);
							continue;
					}
				}
			} catch (GLib.KeyFileError e) {
				GLib.stderr.printf("Problem with configuration in '%s'\n", g);
				continue;
			}
			cfgs.append(cfg);
		}
	} catch (GLib.Error e) {
		GLib.stderr.printf("Error: problem loading config file: %s\n", e.message);
		return null;
	}
	return cfgs;
}


public void config_save() throws GLib.FileError {
    var cfg_path = GLib.Path.build_filename(Environment.get_home_dir(), CONFIG_FILE);
    var cfg_dir = GLib.Path.get_dirname(cfg_path);
    if (!FileUtils.test(cfg_dir, FileTest.IS_DIR))
        DirUtils.create_with_parents(cfg_dir, 0755);
    var buf = "#################### DesktopFolders configuration file ####################\n\n";
    var i = 1;
    foreach (var df in dfs) {
        buf += df.cfg.to_string("FolderDesktop #%d".printf(i));
        i++;
    }
    buf += "#################### End of file ####################";
    FileUtils.set_contents(cfg_path, buf);
}


///////////////////////////////////////////////////////////////////////////
///// MyDialog
public class MyDialog : Gtk.Dialog {
    Gtk.Entry entry;
    string    orig_text;

    public MyDialog(string title, string message, string text) {
        this.title = title;
        this.border_width = 10;
        this.set_default_size(400, 100);
        var vbox = new Gtk.Box(Gtk.Orientation.VERTICAL, 5);
        var hbox = new Gtk.Box(Gtk.Orientation.HORIZONTAL, 15);
        vbox.pack_start(hbox, false, false, 0);
        var img = new Gtk.Image.from_stock(Gtk.Stock.DIALOG_QUESTION, Gtk.IconSize.DIALOG);
        hbox.pack_start(img, false, false, 0);
        var vbox2 = new Gtk.Box(Gtk.Orientation.VERTICAL, 5);
        hbox.pack_start(vbox2, true, true, 0);
        var label = new Gtk.Label("");
        label.set_markup("<b>%s:</b>".printf(message));
        label.set_alignment(0, (float) 0.5);
        vbox2.pack_start(label, false, false, 0);
        this.entry = new Gtk.Entry();
        this.entry.activate.connect(() => { this.response(Gtk.ResponseType.OK); });
        this.orig_text = text;
        this.entry.set_text(text);
        vbox2.pack_start(this.entry, true, true, 0);
        vbox.pack_start(new Gtk.HSeparator(), false, false, 10);
        var content = this.get_content_area() as Box;
        content.pack_start(vbox, true, true, 0);
        this.add_button (Gtk.Stock.CANCEL, Gtk.ResponseType.CANCEL);
        this.add_button (Gtk.Stock.OK, Gtk.ResponseType.OK);
        this.show_all();
    }

    public string run_and_get_text() {
        var ret = this.run();
        var val = this.entry.get_text();
        if (ret!=Gtk.ResponseType.OK || val==this.orig_text || val=="")
            return "";
        return val;
    }
}


///////////////////////////////////////////////////////////////////////////
///// PreferencesDialog
public class PreferencesDialog : Gtk.Dialog {
    public PreferencesDialog(DesktopFolder df) {
        this.title = "Desktop Folder Preferences";
        this.border_width = 10;
        this.set_default_size(400, -1);
        var vbox = new Gtk.Box(Gtk.Orientation.VERTICAL, 10);
        var header = new Gtk.Label("");
        header.set_markup("<span size='large'><b><u>Desktop Folder Preferences</u></b></span>");
        vbox.pack_start(header, false, false, 10);

        var vbox1 = new Gtk.Box(Gtk.Orientation.VERTICAL, 0);
        vbox.pack_start(vbox1, true, false, 0);
        vbox1.pack_start(this.build_section("General"), false, false, 0);
        var align1 = new Gtk.Alignment((float) 0.5, (float) 0.5, (float) 1.0, (float) 1.0);
        align1.left_padding = 6;
        align1.right_padding = 20;
        vbox1.pack_start(align1, true, false, 2);
        var table1 = new Gtk.Table(1, 2, false);
        table1.set_row_spacings(2);
        table1.set_col_spacings(5);
        align1.add(table1);
        table1.attach(this.build_item("Folder path:"), 0, 1, 0, 1, Gtk.AttachOptions.FILL, Gtk.AttachOptions.FILL, 0, 0);
        var fbtn = new Gtk.FileChooserButton("Pick a folder", Gtk.FileChooserAction.SELECT_FOLDER);
        fbtn.set_current_folder_uri(df.cfg.uri);
        fbtn.selection_changed.connect(() => {
                var uri = fbtn.get_current_folder_uri();
                if (uri == null) {
                    fbtn.set_current_folder_uri(df.cfg.uri);
                    return;
                }
                if (df.cfg.uri != uri) {
                    df.cfg.uri = uri;
                    df.update_contents();
                }
            });
        table1.attach_defaults(fbtn, 1, 2, 0, 1);

        var vbox2 = new Gtk.Box(Gtk.Orientation.VERTICAL, 0);
        vbox.pack_start(vbox2, true, false, 5);
        vbox2.pack_start(this.build_section("Filter"), false, false, 0);
        var align21 = new Gtk.Alignment((float) 0.5, (float) 0.5, (float) 1.0, (float) 1.0);
        align21.left_padding = 24;
        align21.right_padding = 20;
        vbox2.pack_start(align21, true, false, 2);
        var vbox21 = new Gtk.Box(Gtk.Orientation.VERTICAL, 0);
        align21.add(vbox21);
        var cbtn_hidden = new Gtk.CheckButton.with_label("Include hidden files");
        cbtn_hidden.active = df.cfg.filter.include_hidden;
        cbtn_hidden.toggled.connect(() => {
                df.cfg.filter.include_hidden = cbtn_hidden.active;
                df.update_contents();
            });
        vbox21.pack_start(cbtn_hidden, true, false, 0);
        var cbtn_backups = new Gtk.CheckButton.with_label("Include backup files");
        cbtn_backups.active = df.cfg.filter.include_backups;
        cbtn_backups.toggled.connect(() => {
                df.cfg.filter.include_backups = cbtn_backups.active;
                df.update_contents();
            });
        vbox21.pack_start(cbtn_backups, true, false, 0);
        var cbtn_folders = new Gtk.CheckButton.with_label("Include folders");
        cbtn_folders.active = df.cfg.filter.include_folders;
        cbtn_folders.toggled.connect(() => {
                df.cfg.filter.include_folders = cbtn_folders.active;
                df.update_contents();
            });
        vbox21.pack_start(cbtn_folders, true, false, 0);
        var cbtn_files = new Gtk.CheckButton.with_label("Include files");
        cbtn_files.active = df.cfg.filter.include_files;
        cbtn_files.toggled.connect(() => {
                df.cfg.filter.include_files = cbtn_files.active;
                df.update_contents();
            });
        vbox21.pack_start(cbtn_files, true, false, 0);
        var align22 = new Gtk.Alignment((float) 0.5, (float) 0.5, (float) 1.0, (float) 1.0);
        align22.left_padding = 6;
        align22.right_padding = 20;
        vbox2.pack_start(align22, true, false, 0);
        var table2 = new Gtk.Table(2, 2, false);
        align22.add(table2);
        table2.set_row_spacings(2);
        table2.set_col_spacings(5);
        table2.attach(this.build_item("Show only list:"), 0, 1, 0, 1, Gtk.AttachOptions.FILL, Gtk.AttachOptions.FILL, 0, 0);
        var entry_showonly = new Gtk.Entry();
        // entry_showonly.secondary_icon_name = "gtk-edit";
        entry_showonly.set_text(string.joinv(",", df.cfg.filter.showonly_list));
        entry_showonly.activate.connect(() => {
                df.cfg.filter.showonly_list = this.string2array(entry_showonly.get_text());
                df.update_contents();
            });
        entry_showonly.focus_out_event.connect(() => {
                df.cfg.filter.showonly_list = this.string2array(entry_showonly.get_text());
                df.update_contents();
                return false;
            });
        // entry_showonly.icon_press.connect((iconpos, ev) => {
        //         if (iconpos!=Gtk.EntryIconPosition.SECONDARY || ev.button.button!=1)
        //             return;
        //         print("open window\n");
        //     });
        table2.attach_defaults(entry_showonly, 1, 2, 0, 1);
        table2.attach(this.build_item("Exclude list:"), 0, 1, 1, 2, Gtk.AttachOptions.FILL, Gtk.AttachOptions.FILL, 0, 0);
        var entry_exclude = new Gtk.Entry();
        // entry_exclude.secondary_icon_name = "gtk-edit";
        entry_exclude.set_text(string.joinv(",", df.cfg.filter.exclude_list));
        entry_exclude.activate.connect(() => {
                df.cfg.filter.exclude_list = this.string2array(entry_exclude.get_text());
                df.update_contents();
            });
        entry_exclude.focus_out_event.connect(() => {
                df.cfg.filter.exclude_list = this.string2array(entry_exclude.get_text());
                df.update_contents();
                return false;
            });
        // entry_exclude.icon_press.connect((iconpos, ev) => {
        //         if (iconpos!=Gtk.EntryIconPosition.SECONDARY || ev.button.button!=1)
        //             return;
        //         print("open window\n");
        //     });
        table2.attach_defaults(entry_exclude, 1, 2, 1, 2);

        var vbox3 = new Gtk.Box(Gtk.Orientation.VERTICAL, 0);
        vbox.pack_start(vbox3, true, false, 0);
        vbox3.pack_start(this.build_section("Look &amp; feel"), false, false, 0);
        var align31 = new Gtk.Alignment((float) 0.5, (float) 0.5, (float) 1.0, (float) 1.0);
        align31.left_padding = 24;
        align31.right_padding = 20;
        vbox3.pack_start(align31, true, false, 2);
        var cbtn_fullpath = new Gtk.CheckButton.with_label("Show full path in header");
        cbtn_fullpath.active = df.cfg.ui_header_fullpath;
        cbtn_fullpath.toggled.connect(() => {
                df.cfg.ui_header_fullpath = cbtn_fullpath.active;
                df.update_contents();
            });
        align31.add(cbtn_fullpath);
        var align32 = new Gtk.Alignment((float) 0.5, (float) 0.5, (float) 1.0, (float) 1.0);
        align32.left_padding = 6;
        align32.right_padding = 20;
        vbox3.pack_start(align32, true, false, 2);
        var table3 = new Gtk.Table(4, 2, false);
        table3.set_row_spacings(2);
        table3.set_col_spacings(5);
        align32.add(table3);
        table3.attach(this.build_item("Path font:"), 0, 1, 0, 1, Gtk.AttachOptions.FILL, Gtk.AttachOptions.FILL, 0, 0);
        var bfont_header = new Gtk.FontButton.with_font(df.cfg.ui_header_font);
        bfont_header.font_set.connect(() => {
                df.cfg.ui_header_font = bfont_header.font_name;
                df.cfg.update_geo();
                df.update_win_geometry();
            });
        table3.attach_defaults(bfont_header, 1, 2, 0, 1);
        table3.attach(this.build_item("Item icon size:"), 0, 1, 1, 2, Gtk.AttachOptions.FILL, Gtk.AttachOptions.FILL, 0, 0);
        var cmb_size = new Gtk.ComboBoxText();
        var active = 0;
        for (int i=0; i<LIST_ICON_SIZE_STR.length; i++) {
            if (df.cfg.ui_icon_size == LIST_ICON_SIZE[i])
                active = i;
            cmb_size.append_text(LIST_ICON_SIZE_STR[i]);
        }
        cmb_size.set_active(active);
        cmb_size.changed.connect(() => {
                var s = cmb_size.get_active_text();
                var index = -1;
                for (int i=0; i<LIST_ICON_SIZE_STR.length; i++) {
                    if (s == LIST_ICON_SIZE_STR[i]) {
                        index = i;
                        break;
                    }
                }
                if (index==-1 || LIST_ICON_SIZE[index]==df.cfg.ui_icon_size)
                    return;
                df.cfg.ui_icon_size = LIST_ICON_SIZE[index];
                df.cfg.update_geo();
                df.update_win_geometry();
            });
        table3.attach_defaults(cmb_size, 1, 2, 1, 2);
        table3.attach(this.build_item("Item label font:"), 0, 1, 2, 3, Gtk.AttachOptions.FILL, Gtk.AttachOptions.FILL, 0, 0);
        var bfont_item = new Gtk.FontButton.with_font(df.cfg.ui_label_font);
        bfont_item.font_set.connect(() => {
                df.cfg.ui_label_font = bfont_item.font_name;
                df.cfg.update_geo();
                df.update_win_geometry();
            });
        table3.attach_defaults(bfont_item, 1, 2, 2, 3);
        table3.attach(this.build_item("Number of columns:"), 0, 1, 3, 4, Gtk.AttachOptions.FILL, Gtk.AttachOptions.FILL, 0, 0);
        var slider_ncols = new Gtk.Scale.with_range(Gtk.Orientation.HORIZONTAL, 1, 20, 1);
        slider_ncols.adjustment.value = df.cfg.ncols;
        slider_ncols.value_changed.connect(() => {
                df.cfg.ncols = (int) slider_ncols.adjustment.value;
                df.cfg.update_geo();
                df.update_win_geometry();
            });
        table3.attach_defaults(slider_ncols, 1, 2, 3, 4);

        vbox.pack_start(new Gtk.HSeparator(), false, false, 10);
        var content = this.get_content_area() as Box;
        content.pack_start(vbox, true, true, 0);
        this.add_button (Gtk.Stock.CLOSE, Gtk.ResponseType.CLOSE);
        this.show_all();
    }

    private Gtk.Label build_section(string text) {
        var label = new Gtk.Label("");
        label.set_alignment(0, (float) 0.5);
        label.set_padding(10, 0);
        label.set_markup("<span size='medium'><b>" + text + "</b></span>");
        return label;
    }

    private Gtk.Label build_item(string text) {
        var label = new Gtk.Label(text);
        label.set_alignment(0, (float) 0.5);
        label.set_padding(20, 0);
        return label;
    }

    private string[] string2array(string text) {
        string[] l = {};
        foreach (var s in text.split(",")) {
            if (s.strip() != "")
                l += s.strip();
        }
        return l;
    }
}


///////////////////////////////////////////////////////////////////////////
///// UI utils
public Cairo.ImageSurface pixbuf2imgsurface(Gdk.Pixbuf pb) {
	var fmt = pb.has_alpha ? Cairo.Format.ARGB32 : Cairo.Format.RGB24;
	var imgsurface = new Cairo.ImageSurface(fmt, pb.width, pb.height);
	var ctx = new Cairo.Context(imgsurface);
	Gdk.cairo_set_source_pixbuf(ctx, pb, 0, 0);
	ctx.paint();
	return imgsurface;
}


// from http://code.valaide.org/content/get-pixbuf-stock
Gdk.Pixbuf get_pixbuf_from_stock (string stock_id, Gtk.IconSize size) {
	var w = new Gtk.Invisible();
	return w.render_icon(stock_id, size, "vala");
}


///////////////////////////////////////////////////////////////////////////
///// DesktopFolder
public class DesktopFolder : Gtk.Window {

    public ArrayList<DFItem> items;
    public DFConfig          cfg;
    ArrayList<int> selection;
    FileMonitor dir_monitor;
    int  cursor_item;
    int  clicked_on;
    bool needs_redraw_canvas;
    bool is_dnd_active;

    enum TargetType { URI_LIST }
    const Gtk.TargetEntry[] DND_TARGETS = {
        { "text/uri-list", Gtk.TargetFlags.OTHER_WIDGET, TargetType.URI_LIST }
    };
    enum ActionType { COPY, MOVE, LINK }

    Gtk.EventBox       evbox;
    Gtk.Menu           menu;
    Gtk.MessageDialog  dlg_error;
    Gtk.MessageDialog  dlg_question;
    Cairo.ImageSurface cache_surface;
    Cairo.ImageSurface img_link;
    Cairo.ImageSurface img_selected;


    ///// initialization
    public DesktopFolder(DFConfig? cfg=null) {
        GLib.Object(type: Gtk.WindowType.TOPLEVEL);
        this.cfg = (cfg==null) ? new DFConfig() : cfg;
        this.evbox = new Gtk.EventBox();
        this.add(this.evbox);
        this.title = PROGRAM_NAME;
        this.set_size_request(2*2 + 2*BORDER + 1*this.cfg.ui_item_width + (1-1)*ITEM_SPACING,
                              2*2 + 2*BORDER + this.cfg.ui_header_font_size + 2*SEP_PADDING + SEP_SIZE + 1*this.cfg.ui_item_height + (1-1)*ITEM_SPACING - 2);
        this.set_default_size(this.cfg.w, this.cfg.h);
        this.move(this.cfg.x0, this.cfg.y0);
        // this.resizable = false;
        this.destroy.connect(() => {
                this.dir_monitor.cancel();
            });
        // make window transparent and draw at back of desktop
        // this.type_hint = Gdk.WindowTypeHint.DOCK;
        this.app_paintable = true;
        // Need to get the rgba colormap or transparency doesn't work.
        this.set_visual(screen.get_rgba_visual());
        this.decorated = false;
        this.skip_taskbar_hint = true;
        this.skip_pager_hint = true;
        this.set_keep_below(true);
        this.stick();
        // add events
        this.evbox.add_events(Gdk.EventMask.BUTTON_PRESS_MASK);
        this.evbox.add_events(Gdk.EventMask.BUTTON_RELEASE_MASK);
        this.evbox.add_events(Gdk.EventMask.POINTER_MOTION_MASK);
        this.evbox.add_events(Gdk.EventMask.LEAVE_NOTIFY_MASK);
        this.evbox.button_release_event.connect(this.on_button_release_cb);
        this.evbox.button_press_event.connect(this.on_button_press_cb);
        this.evbox.motion_notify_event.connect(this.on_motion_cb);
        this.evbox.leave_notify_event.connect(this.on_leave_cb);
        this.evbox.draw.connect(this.on_expose_cb);
        this.configure_event.connect(this.on_resize_cb);
        // drag & drop
        Gtk.drag_source_set(this.evbox, Gdk.ModifierType.BUTTON1_MASK, DND_TARGETS, Gdk.DragAction.COPY|Gdk.DragAction.MOVE|Gdk.DragAction.LINK);
        this.evbox.drag_begin.connect(this.on_drag_begin);
        this.evbox.drag_end.connect(this.on_drag_end);
        this.evbox.drag_data_get.connect(this.on_drag_data_get);
        Gtk.drag_dest_set(this.evbox, Gtk.DestDefaults.ALL, DND_TARGETS, Gdk.DragAction.COPY|Gdk.DragAction.MOVE|Gdk.DragAction.LINK);
        this.evbox.drag_data_received.connect(this.on_drag_data_received);
        this.evbox.drag_motion.connect(() => { this.is_dnd_active = true; return false; });
        this.evbox.drag_leave.connect(() => { this.is_dnd_active = false; });
        // intialize some variables
        this.items = null;
        this.selection = new ArrayList<int>();
        this.cursor_item = -1;
        this.clicked_on = -1;
        this.dir_monitor = null;
        this.needs_redraw_canvas = true;
        this.is_dnd_active = false;
        // initialize ui stuff
        this.img_link = pixbuf2imgsurface(get_pixbuf_from_stock("gtk-undo", Gtk.IconSize.SMALL_TOOLBAR));
        this.img_selected = pixbuf2imgsurface(get_pixbuf_from_stock("gtk-apply", Gtk.IconSize.SMALL_TOOLBAR));
        this.dlg_error = new Gtk.MessageDialog(this,
                                               Gtk.DialogFlags.DESTROY_WITH_PARENT | Gtk.DialogFlags.MODAL,
                                               Gtk.MessageType.ERROR, Gtk.ButtonsType.CLOSE, "");
        this.dlg_question = new Gtk.MessageDialog(this,
                                                  Gtk.DialogFlags.DESTROY_WITH_PARENT | Gtk.DialogFlags.MODAL,
                                                  Gtk.MessageType.QUESTION, Gtk.ButtonsType.YES_NO, "");
        this.update_contents();
        this.show_all();
    }

    private void _build_menu(Gtk.Menu menu) {
		DFItem? it = (this.cursor_item==-1) ? null : this.items[this.cursor_item];
        // open
        var mOpenfolder = new Gtk.ImageMenuItem.from_stock(Gtk.Stock.DIRECTORY, null);
        mOpenfolder.set_label("Open containing folder");
        mOpenfolder.activate.connect(() => {
                var cmd = "xdg-open \"%s\"".printf(this.cfg.uri);
                var errmsg = "<b>OPEN ERROR</b>\n\nCan't open folder:\n<i>%s</i>".printf(this.cfg.uri);
                this.run_command(cmd, errmsg);
            });
        menu.append(mOpenfolder);
        if (it != null) {
            var errmsg = "<b>OPEN ERROR</b>\n\nCan't open file:\n<i>%s</i>".printf(it.name);
            var apps = GLib.AppInfo.get_all_for_type(it.fileinfo.get_content_type());
            var menu2 = new Gtk.Menu();
            foreach (GLib.AppInfo app in apps) {
                var the_app = app; // needed for connect()
                var mApp = new Gtk.ImageMenuItem.with_label(app.get_display_name());
                mApp.set_image(new Gtk.Image.from_gicon(app.get_icon(), Gtk.IconSize.MENU));
                mApp.activate.connect(() => { this.run_with_app(the_app, it, errmsg); });
                menu2.append(mApp);
            }
            var mOpenwith = new Gtk.ImageMenuItem.from_stock(Gtk.Stock.OPEN, null);
            mOpenwith.set_label("Open item with");
            mOpenwith.set_submenu(menu2);
            menu.append(mOpenwith);
        }
        menu.append(new Gtk.SeparatorMenuItem());
        // selection
        if (it != null) {
            var itno = this.cursor_item;
            if (this.cursor_item in this.selection) {
                var mUnselect = new Gtk.ImageMenuItem.from_stock(Gtk.Stock.CANCEL, null);
                mUnselect.set_label("Unselect item");
                mUnselect.activate.connect(() => {
                        this.selection.remove(itno);
                        this.queue_draw();
                    });
                menu.append(mUnselect);
            } else {
                var mSelect = new Gtk.ImageMenuItem.from_stock(Gtk.Stock.APPLY, null);
                mSelect.set_label("Select item");
                mSelect.activate.connect(() => {
                        this.selection.add(itno);
                        this.queue_draw();
                    });
                menu.append(mSelect);
            }
        }
        if (this.selection.size < this.items.size) {
            var mSelectall = new Gtk.ImageMenuItem.from_stock(Gtk.Stock.APPLY, null);
            mSelectall.set_label("Select all");
            mSelectall.activate.connect(() => {
                    this.selection.clear();
                    for (int i=0; i<this.items.size; i++) {
                        this.selection.add(i);
                    }
                    this.queue_draw();
                });
            menu.append(mSelectall);
        }
        if (this.selection.size > 0) {
            var mUnselectall = new Gtk.ImageMenuItem.from_stock(Gtk.Stock.CANCEL, null);
            mUnselectall.set_label("Unselect all");
            mUnselectall.activate.connect(() => {
                    this.selection.clear();
                    this.queue_draw();
                });
            menu.append(mUnselectall);
        }
        menu.append(new Gtk.SeparatorMenuItem());
        // actions: delete, rename, makedir
        if (it != null) {
            var mDelete = new Gtk.ImageMenuItem.from_stock(Gtk.Stock.DELETE, null);
            mDelete.set_label("Move item to Trash");
            mDelete.activate.connect(() => {
                    var text = "<b>MOVE TO TRASH</b>\n\nAre you sure you want to delete '<i>%s</i>'?".printf(it.name);
                    if (!dlg_question_run(text))
                        return;
                    try {
                        var file = File.new_for_uri(GLib.Path.build_filename(it.parent_uri, it.name));
                        if (!file.trash())
                            file.delete();
                    } catch (GLib.Error e) {
                        this.dlg_error_run("<b>MOVE TO TRASH ERROR</b>\n\n<i>%s</i>\n%s".printf(it.name, e.message));
                    }
                });
            menu.append(mDelete);
        }
        if (this.selection.size > 0) {
            var mDelete2 = new Gtk.ImageMenuItem.from_stock(Gtk.Stock.DELETE, null);
            mDelete2.set_label("Move selected items to Trash");
            mDelete2.activate.connect(() => {
                    var text = "<b>MOVE TO TRASH</b>\n\nAre you sure you want to delete %d items?".printf(this.selection.size);
                    if (!dlg_question_run(text))
                        return;
                    foreach (int i in this.selection) {
                        var el = this.items[i];
                        var file = File.new_for_uri(GLib.Path.build_filename(el.parent_uri, el.name));
                        try {
                            if (!file.trash())
                                file.delete();
                        } catch (GLib.Error e) {
                            this.dlg_error_run("<b>MOVE TO TRASH ERROR</b>\n\n<i>%s</i>\n%s".printf(el.name, e.message));
                        }
                    }
                    this.selection.clear();
                    this.queue_draw();
                });
            menu.append(mDelete2);
        }
        if (it != null) {
            var mRename = new Gtk.ImageMenuItem.from_stock(Gtk.Stock.ITALIC, null);
            mRename.set_label("Rename item...");
            mRename.activate.connect(() => {
                    var dlg = new MyDialog("Rename", "Type the new name", it.name);
                    var newname = dlg.run_and_get_text();
                    dlg.destroy();
                    if (newname == "")
                        return;
                    try {
                        var filesrc = File.new_for_uri(GLib.Path.build_filename(it.parent_uri, it.name));
                        var filedest = File.new_for_uri(GLib.Path.build_filename(it.parent_uri, newname));
                        filesrc.move(filedest, GLib.FileCopyFlags.NONE);
                    } catch (GLib.Error e) {
                        this.dlg_error_run("<b>RENAME ERROR</b>\n\n<i>%s</i>\n%s".printf(it.name, e.message));
                    }
                });
            menu.append(mRename);
        }
        var mMakedir = new Gtk.ImageMenuItem.from_stock(Gtk.Stock.DIRECTORY, null);
        mMakedir.set_label("Create new folder...");
        mMakedir.activate.connect(() => {
                var dlg = new MyDialog("Create new folder", "Type folder name", "");
                var name = dlg.run_and_get_text();
                dlg.destroy();
                if (name == "")
                    return;
                var fullname = GLib.Path.build_filename(this.cfg.uri, name);
                try {
                    File.new_for_uri(fullname).make_directory();
                } catch (GLib.Error e) {
                    this.dlg_error_run("<b>CREATE FOLDER ERROR</b>\n\n<i>%s</i>\n%s".printf(fullname, e.message));
                }
            });
        menu.append(mMakedir);
        menu.append(new Gtk.SeparatorMenuItem());
        // global
        var mAddDF = new Gtk.ImageMenuItem.from_stock(Gtk.Stock.ADD, null);
        mAddDF.set_label("Create new Desktop Folder");
        mAddDF.activate.connect(() => {
                dfs.add(new DesktopFolder());
                this.save_config();
            });
        menu.append(mAddDF);
        var mRemoveDF = new Gtk.ImageMenuItem.from_stock(Gtk.Stock.REMOVE, null);
        mRemoveDF.set_label("Remove this Desktop Folder");
        mRemoveDF.activate.connect(() => {
                var msg = (dfs.size==1) ?
                    "<b>Remove Desktop Folder</b>\n\nThis is the last Desktop Folder.\nAre you sure you want to remove it and quit program?" :
                    "<b>Remove Desktop Folder</b>\n\nAre you sure you want to remove this Desktop Folder?";
                if (!dlg_question_run(msg))
                    return;
                dfs.remove(this);
                this.save_config();
                this.destroy();
                if (dfs.size == 0)
                    Gtk.main_quit();
            });
        menu.append(mRemoveDF);
        var mFitwin = new Gtk.ImageMenuItem.from_stock(Gtk.Stock.ZOOM_FIT, null);
        mFitwin.set_label("Fit window dimensions");
        mFitwin.activate.connect(() => { this.do_fit_win(); });
        menu.append(mFitwin);
        var mPrefs = new Gtk.ImageMenuItem.from_stock(Gtk.Stock.PREFERENCES, null);
        mPrefs.activate.connect(() => {
                var dlg = new PreferencesDialog(this);
                dlg.run();
                dlg.destroy();
                this.save_config();
            });
        menu.append(mPrefs);
        menu.append(new Gtk.SeparatorMenuItem());
        var mAbout = new Gtk.ImageMenuItem.from_stock(Gtk.Stock.ABOUT, null);
        mAbout.activate.connect(() => {
                var dlg = new Gtk.AboutDialog();
                dlg.authors = PROGRAM_AUTHORS;
                dlg.program_name = PROGRAM_NAME;
                dlg.version = PROGRAM_VERSION;
                dlg.website = PROGRAM_WEBSITE;
                dlg.website_label = PROGRAM_WEBSITE;
                dlg.copyright = "(C) 2011  Iñigo Serna";
                dlg.license_type = Gtk.License.GPL_3_0;
                dlg.logo = get_pixbuf_from_stock("gtk-home", Gtk.IconSize.DIALOG);
                dlg.run();
                dlg.destroy();
            });
        menu.append(mAbout);
    }


    ///// callbacks
    private bool on_resize_cb(Gdk.EventConfigure ev) {
// print("CONFIGURE: (%d,%d) - (%d,%d)\n", ev.x, ev.y, ev.width, ev.height);
        var n = (ev.width - 2*2 - 2*BORDER + ITEM_SPACING) / (this.cfg.ui_item_width + ITEM_SPACING);
        if (n != this.cfg.ncols) {
            // this.update_win_geometry(n);
            this.compute_win_geometry(n);
            this.needs_redraw_canvas = true;
        }
        this.cfg.x0 = ev.x;
        this.cfg.y0 = ev.y;
        this.cfg.w = ev.width;
        this.cfg.h = ev.height;
        return false;
    }

    private bool on_button_press_cb(Gdk.EventButton ev) {
// print("Clicked on -> %d\n", this.cursor_item);
        if (ev.button == 3) {
            this.menu = new Gtk.Menu();
            this._build_menu(this.menu);
            this.menu.show_all();
            this.menu.popup(null, null, null, 3, ev.time);
            this.get_window().set_cursor(null);
            return true;
        } else {
            if (this.cursor_item == -1)
                return true;
            this.clicked_on = this.cursor_item;
            return false;
        }
    }

    private bool on_button_release_cb(Gdk.EventButton ev) {
        if (this.is_dnd_active)
            return true;
// print("Released on -> %d\n", this.clicked_on);
        if (this.clicked_on == -1) {
            if (ev.button == 1) {
                if (ev.y <= 2+BORDER+this.cfg.ui_header_font_size+SEP_PADDING+SEP_SIZE) {
                    var cmd = "xdg-open \"%s\"".printf(this.cfg.uri);
                    var errmsg = "<b>OPEN ERROR</b>\n\nCan't open folder:\n<i>%s</i>".printf(this.cfg.uri);
                    this.run_command(cmd, errmsg);
                }
            } else if (ev.button == 2) {
                this.do_fit_win();
            }
        } else {
            if (ev.button==1 && (ev.state&Gdk.ModifierType.CONTROL_MASK) != 0) {
                if (this.clicked_on in this.selection)
                    this.selection.remove(this.clicked_on);
                else
                    this.selection.add(this.clicked_on);
                this.queue_draw();
            } else if (ev.button == 1) {
                var it = this.items[this.clicked_on];
                var errmsg = "<b>OPEN ERROR</b>\n\nCan't open file:\n<i>%s</i>".printf(it.name);
                this.run_with_default_app(it, errmsg);
            } else if (ev.button == 2) {
                if (this.clicked_on in this.selection)
                    this.selection.remove(this.clicked_on);
                else
                    this.selection.add(this.clicked_on);
                this.queue_draw();
            }
        }
        this.clicked_on = -1;
        return true;
    }

    private bool on_motion_cb(Gdk.EventMotion ev) {
        var itemno = this._inside_item((int) ev.x, (int) ev.y);
        if (this.cursor_item != itemno) {
            if (itemno == -1) {
                this.get_window().set_cursor(null);
                this.set_tooltip_text("");
            } else {
                this.get_window().set_cursor(new Gdk.Cursor(Gdk.CursorType.HAND2));
                var it = this.items[itemno];
                var name = GLib.Markup.escape_text(it.name);
                var size = format_size(it.fileinfo.get_size());
                var type = it.fileinfo.get_attribute_as_string(GLib.FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE);
                var buf = @"<b>Name:</b>\t$name\n<b>Type:</b>\t$type\n<b>Size:</b>\t$size";
                if (it.islink)
                    buf += "\n<b>Link to:</b>\t" + GLib.Markup.escape_text(it.fileinfo.get_attribute_as_string(GLib.FILE_ATTRIBUTE_STANDARD_SYMLINK_TARGET));
                this.set_tooltip_markup(buf);
            }
// print("Motion -> %d -> % d\n", this.cursor_item, itemno);
            this.cursor_item = itemno;
            this.queue_draw();
        }
        return false;
    }

    private bool on_leave_cb(Gdk.EventCrossing ev) {
        if (this.cursor_item != -1) {
            this.cursor_item = -1;
            this.queue_draw();
        }
        return true;
    }

    private bool on_expose_cb(Cairo.Context ctx) {
        return this.draw_contents(ctx);
    }

    // DnD callbacks
    private void on_drag_begin(Gtk.Widget sender, DragContext context) {
        if (this.is_dnd_active)
            return;
// print("DnD begin -> %d - %d - %d\n", this.cursor_item, this.clicked_on, this.selection.size);
        if (this.selection.size==0 && this.clicked_on==-1)
            return;
        this.is_dnd_active = true;
        if(this.selection.size > 1)
            Gtk.drag_source_set_icon_stock(this.evbox, Gtk.Stock.DND_MULTIPLE);
        else
            // Gtk.drag_source_set_icon_pixbuf(this.evbox, this.items[this.clicked_on].ui.thumb);
            Gtk.drag_source_set_icon_stock(this.evbox, Gtk.Stock.DND);
        this.queue_draw();
    }

    private void on_drag_end(Gtk.Widget sender, Gdk.DragContext context) {
// print("DnD end\n");
        this.is_dnd_active = false;
        this.selection.clear();
        this.clicked_on = -1;
        this.get_window().set_cursor(null);
        this.queue_draw();
    }

    private void on_drag_data_get(Gdk.DragContext context, Gtk.SelectionData selection_data, uint target_type, uint time) {
// print("DnD get start\n");
        if (target_type != TargetType.URI_LIST)
            return;
        string[] uris = {};
        if (this.selection.size > 0) {
            for (int i=0; i<this.selection.size; i++) {
                var it = this.items[this.selection[i]];
                var uri = GLib.Path.build_filename(it.parent_uri, it.name);
                uris += uri;
            }
        } else {
            var it = this.items[this.clicked_on];
            var uri = GLib.Path.build_filename(it.parent_uri, it.name);
            uris += uri;
        }
// foreach (var uri in uris) { print("DnD get: %s\n", uri); }
        selection_data.set_uris(uris);
// print("DnD get end\n");
    }

    private void on_drag_data_received(Gdk.DragContext context, int x, int y, Gtk.SelectionData selection_data, uint target_type, uint time) {
        ActionType action;
        string action_str;
        if (target_type != TargetType.URI_LIST) {
            Gtk.drag_finish(context, false, false, time);
            this.is_dnd_active = false;
            return;
        }
        if (context.get_suggested_action() == Gdk.DragAction.COPY) {
            action = ActionType.COPY;
            action_str = "copying file";
        } else if (context.get_suggested_action() == Gdk.DragAction.MOVE) {
            action = ActionType.MOVE;
            action_str = "moving file";
        } else if (context.get_suggested_action() == Gdk.DragAction.LINK) {
            action = ActionType.LINK;
            action_str = "creating link";
        } else {
            Gtk.drag_finish(context, false, false, time);
            this.is_dnd_active = false;
            return;
        }
        // drop into a folder?
        var dir = "";
        foreach (DFItem it in this.items) {
            if (it.ui.is_inside(x, y)) {
                if (it.isdir)
                    dir = it.name;
                break;
            }
        }
        // copy, move or link
        foreach (string uri in selection_data.get_uris()) {
            var filesrc = File.new_for_uri(uri);
            var filedest = File.new_for_uri(GLib.Path.build_filename(this.cfg.uri, dir, filesrc.get_basename()));
            if (!filesrc.query_exists()) {
                this.dlg_error_run("<b>ERROR</b>\n\nFile does not exists:\n<i>%s</i>".printf(filesrc.get_uri()));
                continue;
            }
            if (action==ActionType.LINK && !filesrc.is_native()) {
                this.dlg_error_run("<b>ERROR</b>\n\nCan't create link to a remote file\n<i>%s</i>".printf(filesrc.get_uri()));
                continue;
            }
            try {
                if (action == ActionType.COPY) {
                    filesrc.copy(filedest, GLib.FileCopyFlags.BACKUP|GLib.FileCopyFlags.NOFOLLOW_SYMLINKS|GLib.FileCopyFlags.ALL_METADATA);
                } else if (action == ActionType.MOVE) {
                    filesrc.move(filedest, GLib.FileCopyFlags.BACKUP|GLib.FileCopyFlags.NOFOLLOW_SYMLINKS|GLib.FileCopyFlags.ALL_METADATA);
                } else if (action == ActionType.LINK) {
                    filedest.make_symbolic_link(filesrc.get_path());
                } else {
                    Gtk.drag_finish(context, false, false, time);
                    this.is_dnd_active = false;
                    return;
                }
            } catch (GLib.Error e) {
                this.dlg_error_run("<b>ERROR %s</b>\n\n%s".printf(action_str, e.message));
            }
        }
        Gtk.drag_finish(context, true, true, time);
        this.is_dnd_active = false;
        this.update_contents(); // sometimes dir monitoring is too slow and don't catch file drop
    }


    ///// drawing
    public void redraw() {
        this.queue_draw();
    }

    private bool draw_contents(Cairo.Context ctx) {
        int w, h;
        this.get_size(out w, out h);
        if (this.cache_surface==null)
            this.cache_surface = new Cairo.ImageSurface(Cairo.Format.ARGB32, w, h);
        if (this.needs_redraw_canvas) {
            this._cache_draw(w, h);
            this.needs_redraw_canvas = false;
        }
        ctx.set_source_rgba(1.0, 1.0, 1.0, 0);
        ctx.set_operator(Cairo.Operator.SOURCE);
        ctx.paint();
        if (this.is_dnd_active) {
            ctx.set_line_width(4);
            ctx.set_source_rgb(0.8, 0.2, 0.2);
            this._draw_rounded_rectangle(ctx, 0, 0, w, h, 15);
            ctx.stroke();
        }
        ctx.save();
        ctx.set_operator(Cairo.Operator.ADD);
        ctx.set_source_surface(this.cache_surface, 0, 0);
        ctx.paint();
        for (int i=0; i<this.items.size; i++) {
            var it = this.items[i];
            if (this.cursor_item == i)
                this._draw_cursor(ctx, it.ui.x0, it.ui.y0, this.cfg.ui_item_width, this.cfg.ui_item_height);
            if (it.islink) {
                ctx.set_source_surface(this.img_link, it.ui.x1-22, it.ui.y0);
                ctx.paint();
            }
            if (i in this.selection) {
                ctx.set_source_surface(this.img_selected, it.ui.x0, it.ui.y0);
                ctx.paint();
            }
        }
        ctx.restore();
        return true;
    }

    private void _cache_draw(int w, int h) {
        this.cache_surface = new Cairo.ImageSurface(Cairo.Format.ARGB32, w, h);
        var ctx = new Cairo.Context(this.cache_surface);
        // rectangle
        ctx.set_line_width(2);
        this._draw_rounded_rectangle(ctx, 0, 0, w, h, 15);
        ctx.set_source_rgba(0, 0, 0, 0.5);
        ctx.fill_preserve();
        ctx.set_source_rgba(0.3, 0.3, 0.3, 0.8);
        ctx.stroke();
        // header
        ctx.move_to(BORDER, BORDER+this.cfg.ui_header_font_size);
        ctx.set_source_rgba(1, 1, 1, 0.9);
        ctx.select_font_face(this.cfg.ui_header_font_name,
                             this.cfg.ui_header_font_italic ? Cairo.FontSlant.ITALIC : Cairo.FontSlant.NORMAL,
                             this.cfg.ui_header_font_bold ? Cairo.FontWeight.BOLD : Cairo.FontWeight.NORMAL);
        ctx.set_font_size(this.cfg.ui_header_font_size);
        ctx.show_text(this.cfg.ui_header_fullpath ? this.cfg.uri : GLib.Path.get_basename(this.cfg.uri));
        ctx.set_line_width(SEP_SIZE);
        ctx.set_source_rgb(0.8, 0.8, 0.8);
        ctx.move_to(BORDER, BORDER+this.cfg.ui_header_font_size+SEP_PADDING);
        ctx.line_to(w-BORDER, BORDER+this.cfg.ui_header_font_size+SEP_PADDING);
        ctx.stroke();
        // contents
        if (this.items.size==0 || this.cfg.ncols==0)
            return;
        ctx.set_source_rgb(0.8, 0.8, 0.8);
        ctx.select_font_face(this.cfg.ui_label_font_name,
                             this.cfg.ui_label_font_italic ? Cairo.FontSlant.ITALIC : Cairo.FontSlant.NORMAL,
                             this.cfg.ui_label_font_bold ? Cairo.FontWeight.BOLD : Cairo.FontWeight.NORMAL);
        ctx.set_font_size(this.cfg.ui_label_font_size);
        foreach (var it in this.items) {
            if (it.ui.thumb != null) {
                Gdk.cairo_set_source_pixbuf(ctx, it.ui.thumb, it.ui.thumb_x, it.ui.thumb_y);
                ctx.paint();
            }
            string label;
            int xpad;
            this._clip_text(ctx, it.name, this.cfg.ui_item_width-4, out label, out xpad);
            it.ui.fill_extra(label, it.ui.x0+2+xpad);
            ctx.set_source_rgb(0.9, 0.9, 0.9);
            ctx.move_to(it.ui.label_x, it.ui.label_y);
            ctx.show_text(it.ui.label);
        }
    }

    private void _draw_rounded_rectangle(Cairo.Context ctx, int x, int y, int w, int h, int r) {
        ctx.move_to(x+r,y);
        ctx.line_to(x+w-r,y);
        ctx.curve_to(x+w,y,x+w,y,x+w,y+r);
        ctx.line_to(x+w,y+h-r);
        ctx.curve_to(x+w,y+h,x+w,y+h,x+w-r,y+h);
        ctx.line_to(x+r,y+h);
        ctx.curve_to(x,y+h,x,y+h,x,y+h-r);
        ctx.line_to(x,y+r);
        ctx.curve_to(x,y,x,y,x+r,y);
    }

    private void _draw_cursor(Cairo.Context ctx, int x, int y, int w, int h) {
        ctx.set_line_width(1);
        ctx.set_source_rgba(0, 0, 0.1, 0.2);
        ctx.rectangle(x, y, w, h);
        ctx.fill_preserve();
        ctx.set_source_rgb(1, 1, 1);
        ctx.stroke();
    }

    private void _clip_text(Cairo.Context ctx, string text, int maxw, out string clipped, out int xpad) {
        Cairo. TextExtents cte;
        ctx.text_extents(text, out cte);
        if (maxw >= cte.width) {
            clipped = text;
        } else {
            var n = text.length*maxw/cte.width - 2;
            clipped = text[0:(long) n] + "\xe2\x80\xa6"; // utf8 for "..."
            ctx.text_extents(clipped, out cte);
        }
        xpad = (maxw-(int) cte.width) / 2;
    }


    ///// UI geometry
    public void update_win_geometry(int ncols=-1) {
        if (ncols == -1)
            this.compute_win_geometry(this.cfg.ncols);
        else
            this.compute_win_geometry(ncols);
        this.do_fit_win(); // does window drawing too
    }

    private void compute_win_geometry(int ncols) {
        this.cfg.ncols = ncols;
        if (this.cfg.ncols == 0) {
            this.cfg.nrows = 0;
            return;
        }
        this.cfg.nrows = (int) this.items.size/ncols + (this.items.size%ncols==0 ? 0 : 1);
        this._compute_items_pos();
        this.set_size_request(2*2 + 2*BORDER + 1*this.cfg.ui_item_width + (1-1)*ITEM_SPACING,
                              2*2 + 2*BORDER + this.cfg.ui_header_font_size + 2*SEP_PADDING + SEP_SIZE + 1*this.cfg.ui_item_height + (1-1)*ITEM_SPACING - 2);
    }

    private void _compute_items_pos() {
        int r, c, x, y, x0, y0;
        x = 2 + BORDER + ITEM_SPACING/2;
        y = 2 + BORDER + this.cfg.ui_header_font_size + SEP_PADDING + SEP_SIZE  + SEP_PADDING + ITEM_SPACING/2;
        for (int i=0; i<this.items.size; i++) {
            r = i / this.cfg.ncols;
            c = i % this.cfg.ncols;
            x0 = x + c*(this.cfg.ui_item_width+ITEM_SPACING);
            y0 = y + r*(this.cfg.ui_item_height+ITEM_SPACING);
            this.items[i].ui.fill_geometry(x0, y0, x0+this.cfg.ui_item_width, y0+this.cfg.ui_item_height, r, c, this.cfg);
        }
    }

    private int _inside_item(int x, int y) {
        for (int i=0; i<this.items.size; i++) {
            if (this.items[i].ui.is_inside(x, y))
                return i;
        }
        return -1;
    }

    private void do_fit_win() {
        var w = 2*2 + 2*BORDER + this.cfg.ncols*this.cfg.ui_item_width + (this.cfg.ncols-1)*ITEM_SPACING;
        var h = 2*2 + 2*BORDER + this.cfg.ui_header_font_size + 2*SEP_PADDING + SEP_SIZE + this.cfg.nrows*this.cfg.ui_item_height + (this.cfg.nrows-1)*ITEM_SPACING;
        this.resize(w, h);
        this.needs_redraw_canvas = true;
        this.queue_draw();
    }


    ///// utils
    private void dlg_error_run(string text) {
        dlg_error.set_markup(text);
        dlg_error.show();
        dlg_error.run();
        dlg_error.hide();
    }

    private bool dlg_question_run(string text) {
        dlg_question.set_markup(text);
        dlg_question.show();
        var res = dlg_question.run() == Gtk.ResponseType.YES;
        dlg_question.hide();
        return res;
    }

    private void run_command(string cmd, string errmsg) {
        this.get_window().set_cursor(new Gdk.Cursor(Gdk.CursorType.WATCH));
        try {
            GLib.Process.spawn_command_line_async(cmd);
        } catch (GLib.SpawnError e) {
            this.dlg_error_run(errmsg);
        }
        GLib.Timeout.add(1000, () => { this.get_window().set_cursor(null); return false; });
    }

    private void run_with_default_app(DFItem it, string errmsg) {
        var app = GLib.AppInfo.get_default_for_type(it.fileinfo.get_content_type(), false);
        this.run_with_app(app, it, errmsg);
    }

    private void run_with_app(GLib.AppInfo app, DFItem it, string errmsg) {
        this.get_window().set_cursor(new Gdk.Cursor(Gdk.CursorType.WATCH));
        try {
            var flist = new GLib.List<File>();
            flist.append(File.new_for_uri(GLib.Path.build_filename(it.parent_uri, it.name)));
            var ok = app.launch(flist, null);
            if (!ok)
                throw new GLib.SpawnError.FAILED(errmsg);
        } catch (GLib.Error e) {
            this.dlg_error_run(errmsg);
        }
        GLib.Timeout.add(1000, () => { this.get_window().set_cursor(null); return false; });
    }


    ///// main
    private void save_config() {
        try {
            config_save();
        } catch (GLib.FileError e) {
            this.dlg_error_run("<b>ERROR SAVING CONFIGURATION</b>\n\n>%s".printf(e.message));
        }
    }

    public void update_contents() {
        if (this.cfg.uri == "")
            this.cfg.uri = "file://" + Environment.get_home_dir();
        var furi = File.new_for_uri(this.cfg.uri);
        if (!furi.query_exists()) {
            this.dlg_error_run("<b>ERROR</b>\n\nCan't load uri: <i>%s</i>\nWill continue with HOME folder".printf(this.cfg.uri));
            this.cfg.uri = "file://" + Environment.get_home_dir();
            furi = File.new_for_uri(this.cfg.uri);
        }
        if (this.items != null)
            this.items.clear();
        if (this.selection != null)
            this.selection.clear();
        try {
            this.items = get_items(furi, this.cfg.filter);
        } catch (GLib.Error e) {
            this.dlg_error_run("<b>ERROR</b>\n\n%s".printf(e.message));
        }
        if (this.dir_monitor != null)
            this.dir_monitor.cancel();
        try {
            this.dir_monitor = furi.monitor_directory(GLib.FileMonitorFlags.SEND_MOVED);
        } catch (GLib.IOError e) {}
        this.dir_monitor.changed.connect((mon, f1, f2, ev) => {
                var file = (f2==null) ? f1 : f2;
                FileInfo fi;
                try {
                    fi = file.query_info("*", FileQueryInfoFlags.NONE);
                } catch (GLib.Error e) {
// print("DIR MONITOR: %s\n", file.get_uri());
                    this.update_contents();
                    return;
                }
                if (this.cfg.filter.is_ok(fi)) {
// print("DIR MONITOR: %s\n", file.get_uri());
                    this.update_contents();
                }
            });
        this.update_win_geometry(this.cfg.ncols); // does window drawing too
    }
}


///////////////////////////////////////////////////////////////////////////
///// Main
public static int main(string[] args) {
    Gtk.init(ref args);
    // initialize thumb factory and build instances list
    thumb_factory = new ThumbFactory();
    dfs = new ArrayList<DesktopFolder>();
    // load configuration or clean start
    var cfgs = config_load();
    if (cfgs==null || cfgs.length()==0) {
        dfs.add(new DesktopFolder(new DFConfig()));
    } else {
        foreach (var cfg in cfgs) {
            dfs.add(new DesktopFolder(cfg));
        }
    }
    // timeout for saving configuration
    GLib.Timeout.add_seconds(CONFIG_SAVING_TIMEOUT, () => {
            try {
                config_save();
            } catch (GLib.FileError e) {}
            return true;
        });
    Gtk.main();
    return 0;
}


///////////////////////////////////////////////////////////////////////////
